// generated by verigenSV.pl on Fri Jun  4 12:21:08 2021
// input file: camt8d32.cfg

// compiler directives
/* verilator lint_off CASEINCOMPLETE */
/* verilator lint_off DECLFILENAME */
/* verilator lint_off UNUSED */
/* verilator lint_off STMTDLY */
/* verilator lint_off UNOPTFLAT */


module CAMT8D32 (
	input logic clk,
	input logic res_n,
	input logic lock,
	input logic unlock,
	input logic [CAM_KEY_WIDTH-1:0] key,
	input logic wr,
	input logic [CAM_WIDTH-1:0] din,
	input logic rd,
	output logic dout_val, 
	output logic [CAM_WIDTH-1:0] dout, 
	output logic locked );

  //
 // parameter constants 
//---------------------------
parameter int CAM_SIZE = 16;
parameter int CAM_WIDTH = 40;
parameter int CAM_TAG_WIDTH = 8;
parameter int CAM_KEY_WIDTH = 16;
parameter logic [CAM_KEY_WIDTH-1:0] KEY = 16'hBABE;
parameter logic [CAM_WIDTH-1:0] ZERO_DATA = {CAM_WIDTH{1'b0}};


  //
 // state encoding typedef 
//--------------------------------
typedef enum logic [0:0] {
	LOCKED = 1'b1,
	UNLOCKED = 1'b0
} states_t;


// internal state variables
states_t STATE, NEXT_STATE;

  //
 // memory instantiations 
//---------------------------
uwire T8D32_CAM_hit;
CAM #( CAM_WIDTH, CAM_SIZE, CAM_TAG_WIDTH ) T8D32_CAM(
	.clk(clk), .res_n(res_n), 
	.wr(T8D32_CAM_wr), .wr_data(T8D32_CAM_wr_data[CAM_WIDTH-1:0]), 
	.rd(T8D32_CAM_rd), .rd_data(T8D32_CAM_rd_data[CAM_WIDTH-1:0]), 
	.hit(T8D32_CAM_hit) );

   //
  // implicit continuous assignments 
 // (derived from <expressions> section) 
//----------------------------------
uwire DO_IDLE; assign DO_IDLE = !lock & !unlock;
uwire DO_LOCK ; assign DO_LOCK  =  lock & !unlock ;
uwire DO_READ_CAM; assign DO_READ_CAM = (UNLOCKED==STATE) & rd & !wr;
uwire DO_UNLOCK; assign DO_UNLOCK = !lock &  unlock & IS_KEY_MATCH ;
uwire DO_WRITE_CAM; assign DO_WRITE_CAM = (UNLOCKED==STATE) & wr & !rd;
uwire IS_KEY_MATCH; assign IS_KEY_MATCH = (KEY==key);

  //
 // initials for simulator
//---------------------------
initial begin
  STATE = LOCKED;
  T8D32_CAM_wr = 1'b0;
  T8D32_CAM_rd = 1'b0;
  T8D32_CAM_wr_data[CAM_WIDTH-1:0] = {CAM_WIDTH{1'b0}};

  // dump output (Verilator)
  if ($test$plusargs("trace") != 0) begin
    $display("[%0t] Tracing to logs/vlt_dump.vcd...\n", $time);
    $dumpfile("logs/vlt_dump.vcd");
    $dumpvars();
  end
  $display("[%0t] Model running...\n", $time);

end

  //
 // state transfer
//---------------------------
always_ff @( posedge clk ) begin
  if( res_n )
    STATE <= NEXT_STATE;
  else
    STATE <= LOCKED;

  // condition for $finish (Verilator)
  if( lock==1'b1 && unlock==1'b1 && rd==1'b1 && wr==1'b1 && res_n==1'b0 ) begin
    $write("*-* All Finished *-*\n");
    $finish;
  end   

end

// AssertionExample:
//always_ff @ (posedge clk) begin
//   AssertionExample: assert (rd==1'b1 ^ wr==1'b1);
//end

// coverage analysis (Verilator)
// cover property ( @(posedge clk) rd_q0==1'b1 && rd_q1==1'b1 && res_n==1'b0 );

always_comb begin

    //
   // pre-case (default) assignments 
  //----------------------------------
  dout = ZERO_DATA;
  dout_val = 1'b0;
  locked = STATE;
  NOP_T8D32_CAM();

  unique case ( STATE )
  LOCKED : begin
	if ( DO_UNLOCK ) begin
	  NEXT_STATE = UNLOCKED;
	end
	if ( !DO_UNLOCK ) begin
	  NEXT_STATE = LOCKED;
	end
  end // LOCKED
  UNLOCKED : begin
	if ( DO_IDLE ) begin
	  NEXT_STATE = UNLOCKED;
	  PUSH_T8D32_CAM(DO_WRITE_CAM,din);
	  dout = POP_T8D32_CAM(DO_READ_CAM,din);
	  dout_val=T8D32_CAM_hit;
	end
	if ( DO_LOCK ) begin
	  NEXT_STATE = LOCKED;
	end
  end // UNLOCKED
  endcase
end // always_comb


  //
 // memory functions
//---------------------------
logic T8D32_CAM_rd;
logic [CAM_WIDTH-1:0] T8D32_CAM_rd_data;

function automatic [CAM_WIDTH-1:0] POP_T8D32_CAM( read, [CAM_WIDTH-1:0] read_data );
   T8D32_CAM_rd = read;
   T8D32_CAM_wr_data = read_data;
   return T8D32_CAM_rd_data;
endfunction

logic T8D32_CAM_wr;
logic [CAM_WIDTH-1:0] T8D32_CAM_wr_data;

function automatic void PUSH_T8D32_CAM( write, [CAM_WIDTH-1:0] write_data );
   T8D32_CAM_wr = write;
   T8D32_CAM_wr_data = write_data;
endfunction

function automatic void NOP_T8D32_CAM();
   T8D32_CAM_rd = 1'b0;
   T8D32_CAM_wr = 1'b0;
   T8D32_CAM_wr_data = '0;
   T8D32_CAM_rd_data = '0;
endfunction

endmodule


  //
 // memory
//---------------------------
module CAM ( clk, res_n, wr, wr_data, rd, rd_data, hit );

parameter DATA_WIDTH = 32;
parameter NUM_ENTRIES = 16;
parameter TAG_WIDTH = 8;

input  logic clk, res_n, rd, wr;
input  logic [DATA_WIDTH-1:0] wr_data;
output logic [DATA_WIDTH-1:0] rd_data;
output logic hit;
    
localparam integer ADDR_WIDTH = $clog2(NUM_ENTRIES);
localparam integer ENTRY_WIDTH = DATA_WIDTH - TAG_WIDTH;

localparam [ADDR_WIDTH-1:0] ZERO_ADDR = {ADDR_WIDTH{1'b0}};
localparam [DATA_WIDTH-1:0] ZERO_DATA = {DATA_WIDTH{1'b0}};

logic [DATA_WIDTH-1:0] MEMORY [NUM_ENTRIES-1:0];
logic [ADDR_WIDTH-1:0] HEAD, HEAD_PLUS_1; 
logic [NUM_ENTRIES-1:0] compare;

initial begin
    for ( int i = 0; i < NUM_ENTRIES; i = i + 1 ) begin
        MEMORY[ i ] = ZERO_DATA;
    end
    HEAD = ZERO_ADDR;
end

// content addressable compare function
assign compare = ( rd && !wr && res_n ) ? compare_fcn() : '0;

// in-cycle read access
assign rd_data = ( rd && !wr && res_n ) ? read_fcn() : ZERO_DATA;

// append entries (wrap)
assign HEAD_PLUS_1 = HEAD + 1'b1;

always @ ( posedge clk ) begin

  if( res_n ) begin
    if( wr && !rd ) begin // WRITE ACCESS
      MEMORY[HEAD] <= wr_data;
      HEAD <= HEAD_PLUS_1;
    end
  end else begin
      HEAD <= clear_fcn();
  end
end

function automatic [NUM_ENTRIES-1:0] compare_fcn();
  for( int i=0; i<NUM_ENTRIES; i++ ) begin
    compare_fcn[i] = ( MEMORY[i][DATA_WIDTH-1:ENTRY_WIDTH] == wr_data[DATA_WIDTH-1:ENTRY_WIDTH] );
  end
endfunction

function automatic [DATA_WIDTH-1:0] read_fcn();
  read_fcn = ZERO_DATA;
  hit = 1'b0;
  for( int i=0; i<NUM_ENTRIES; i++ )
    if( 1'b1 == compare[i] ) begin
      read_fcn = MEMORY[i];
      hit = rd && !wr && compare[i];
      break;	
    end
endfunction

function automatic [ADDR_WIDTH-1:0] clear_fcn();
  for( int i=0; i<NUM_ENTRIES; i++ )
      MEMORY[ i ] <= ZERO_DATA;
  clear_fcn = ZERO_ADDR;
endfunction

endmodule
